from datetime import datetime
from typing import List, Dict, Any
from sqlalchemy import Engine, select, create_engine
from sqlalchemy.orm import Session
from .models import Tile, SentinelImage, MissingImage, ImageToProcess


def init_db_engine(user: str, password: str, host: str, database: str) -> Engine:
    """Initializes sqlalchemy engine that connects to postgis database

    Args:
        user (str): the user name
        password (str): the password
        host (str): the url of the database host
        database (str): the name of the database

    Returns:
        Engine: connection to postgis database
    """

    return create_engine(f"postgresql+psycopg2://{user}:{password}@{host}/{database}")


def fetch_image_data(tile_name: str, engine: Engine):
    with Session(engine) as session:
        result = []
        requested = session.scalars(select(Tile).where(Tile.name == tile_name))
        for tile in requested:
            result.append(
                {"id": tile.id, "last_img_uuid": tile.last_img_uuid, "name": tile.name}
            )

        return result


def fetch_requested_images(engine: Engine) -> List[Dict[str, Any]]:
    with Session(engine) as session:
        result = []
        requested = session.scalars(
            select(Tile).where(Tile.last_img_status == "requested")
        )
        for tile in requested:
            result.append(
                {
                    "id": tile.id,
                    "last_img_uuid": tile.last_img_uuid,
                    "name": tile.name,
                }
            )

        return result


def fetch_required_tiles(engine: Engine) -> List[Tile]:
    """Fetches tiles that completely cover Chile (i.e. required=true)"""
    with Session(engine) as session:
        result = []
        query = select(Tile).where(Tile.required).order_by(Tile.name)
        for tile in session.scalars(query):
            result.append(tile)
        return result


def fetch_missing_images(engine: Engine) -> List[ImageToProcess]:
    """Fetches images that have not been processed"""
    with Session(engine) as session:
        result = []
        img_query = select(Tile.id, MissingImage.tile, MissingImage.uuid).join(
            Tile, onclause=Tile.name == MissingImage.tile
        )
        for img in session.execute(img_query):
            result.append(ImageToProcess(img[0], img[1], img[2]))
        return result


def fetch_tile_images(
    engine: Engine, tile: str, columns: List = [SentinelImage]
) -> List[datetime]:
    result = []
    query = select(*columns).where(
        SentinelImage.required == True, SentinelImage.tile == tile
    )
    with Session(engine) as session:
        for row in session.execute(query):
            result.append(row)
    return result


def add_missing_image(engine: Engine, uuid: str, tile: str):
    with Session(engine) as session:
        new_image = MissingImage(uuid=uuid, tile=tile)
        session.add(new_image)



""" Models for this app """
from datetime import datetime
from typing import Optional

from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column
from sqlalchemy import DateTime
from geoalchemy2 import Geometry


class Base(DeclarativeBase):
    """Base Class of the model"""


class Tile(Base):
    """Tile model"""

    __tablename__ = "tiles"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    geom = mapped_column(Geometry("POLYGON", srid=4326))
    last_date: Mapped[Optional[datetime]] = mapped_column(DateTime(timezone=True))
    last_img_status: Mapped[str] = mapped_column(default="none")
    last_s3_key: Mapped[Optional[str]]
    required: Mapped[bool]
    last_img_uuid: Mapped[Optional[str]]


class SentinelImage(Base):
    """Sentinel Image"""

    __tablename__ = "sentinel_images"
    id: Mapped[int] = mapped_column(primary_key=True)
    uuid: Mapped[str]
    tile: Mapped[str]
    date: Mapped[datetime] = mapped_column(DateTime(timezone=True))
    s3_key: Mapped[str]
    cloud: Mapped[float]
    footprint = mapped_column(Geometry("POLYGON", srid=4326))
    required: Mapped[bool] = mapped_column(default=False)


class MissingImage(Base):
    """Missing Sentinel Image"""

    __tablename__ = "missing_images"
    uuid: Mapped[str] = mapped_column(primary_key=True)
    tile: Mapped[str]


class ImageToProcess:
    """Helper model for keeping track of images to process"""

    def __init__(self, tile_id: int, tile: str, uuid: str):
        self.tile = tile
        self.uuid = uuid
        self.tile_id = tile_id
